/*
  Last changed Time-stamp: <2008-06-18 13:07:00 xtof>
  $Id$
  -*- mode: c++; c-indent-level: 2; -*-
*/

//--------------------------------------------------------------------------------------------------------
// includes
//--------------------------------------------------------------------------------------------------------
#include"definitions.h"
#include"decomp_cartSk.h"

#include <iostream>
#include <fstream>


#include <boost/config.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/graph_utility.hpp>
#include <boost/graph/connected_components.hpp>
#include <boost/graph/breadth_first_search.hpp>

//#include <math.h>

//--------------------------------------------------------------------------------------------------------
// namespaces
//--------------------------------------------------------------------------------------------------------
using namespace std;
using namespace boost;


//--------------------------------------------------------------------------------------------------------
// edges to vertices
// input:  cpgraph - cartesian product
//         graph2 - empty graph
//         edgeVector - empty vector
// output: graph2 - new graph with 'n' vertices of 'n' edges of cpgraph
//         edgeVector - stores edges (myEdge) of cpgraph
//--------------------------------------------------------------------------------------------------------
void edgesToVertices(myGraph &cpgraph, myGraph &graph2, vector<myEdge> &edgeVector)
//--------------------------------------------------------------------------------------------------------
{ 
	//myEdgeID  edge_id = get(edge_index_t, cpgraph);
	myEdgeID  Gedge_id = get(edge_index, cpgraph);
	eIt ei, ei_end;
	for(tie(ei, ei_end) = edges(cpgraph); ei != ei_end; ++ei)
	{
		myEdge edge(source(*ei, cpgraph), target(*ei, cpgraph));
		edgeVector.push_back(edge);	
		add_vertex(graph2);
	}
}


//-----------------------------------------------------------------------------------------------------------
//BFSDistance - computes distances from all vertices to the given one
//input: v - given vertex
//       d - array to store distances
//       G - graph
//output: d - stores computed distances
//-----------------------------------------------------------------------------------------------------------
void BFSDistance(vD v, boost::graph_traits<myGraph>::vertices_size_type* d, const myGraph &G)
//-------------------------------------------------------------------------------------------------------------
{
     int n = num_vertices(G);
     std::fill_n(d, n, 0);
     
     boost::breadth_first_search
         (G, v, 
         boost::visitor(boost::make_bfs_visitor
         ((boost::record_distances(d, boost::on_tree_edge())
                         ))) );
}

//--------------------------------------------------------------------------------------------------------
//BFSTree - create BFS tree for a given graph
// input: G - given graph
//        T - graph with the same number of vertices as G, withour any adges
// output: T - spanning tree
//----------------------------------------------------------------------------------------
void BFSTree(myGraph &G, myGraph &T)
//----------------------------------------------------------------------------------------
{
     int n = boost::num_vertices(G);
     int s = *(boost::vertices(G).first); //source vertex
     typedef boost::graph_traits<myGraph>::vertices_size_type vertex_size_t;
     vertex_size_t *d = new vertex_size_t[n]; //array for distances from s
    
     BFSDistance(s, d, G);
     
     myEdgeID  edge_id = get(edge_index, G);
     
     vIt vi, viEnd;
     aIt ai, aiEnd;
     
     for ( tie(vi, viEnd) = vertices(G); vi != viEnd; ++vi){
		 for ( tie(ai, aiEnd) = adjacent_vertices(*vi, G); ai != aiEnd; ++ai){ 
			 if ( d[*vi] == d[*ai] + 1)
			 { 
				eD e = edge(*vi, *ai, G).first;
				add_edge(*vi, *ai, get(edge_id, e), T);
				break;
			 }
		 }
	 }
     
     delete [] d;
}

//--------------------------------------------------------------------------------------
// Theta1 - computes the Theta1 relation and uses it to compute djoko_winkler rel.
// input: G - given graph
//        graph2 - vertices of graph2 correspond with edges of G
// output: graph2 - graph that represents equivalence classes of relation
//--------------------------------------------------------------------------------------
void RELATION_djoko_winkler(myGraph &G, myGraph &graph2)
//--------------------------------------------------------------------------------------
{
        myGraph T(num_vertices(G));
        BFSTree(G, T);
        int n = num_vertices(G);
               
        eIt Tei, TeiEnd;
        
        myEdgeID  Gedge_id = get(edge_index, G);
        myEdgeID  Tedge_id = get(edge_index, T);
        
        for (tie(Tei, TeiEnd) = edges(T); Tei != TeiEnd; ++Tei)
        {
            int v1 = source(*Tei, T);
            int v2 = target(*Tei, T);
	    
            typedef boost::graph_traits<myGraph>::vertices_size_type vertex_size_t;
            vertex_size_t *d1 = new vertex_size_t[n];
            vertex_size_t *d2 = new vertex_size_t[n];
            
            BFSDistance(v1, d1, G);
            BFSDistance(v2, d2, G);
            
            eIt Gei, GeiEnd;
            for (tie(Gei, GeiEnd) = edges(G); Gei != GeiEnd; ++Gei)
            {
                int u1 = source(*Gei, G);
                int u2 = target(*Gei, G);
               
                if ( ( d1[u1] + d2[u2] != d2[u1] + d1[u2])
                     && ( get(Tedge_id, *Tei) != get(Gedge_id, *Gei) )
                     /*&& (! edge(get(Tedge_id, *Tei), get(Gedge_id, *Gei), graph2).second ) */)
                    {
                         add_edge( get(Tedge_id, *Tei), get(Gedge_id, *Gei), graph2);
                    }
            }
	    
	    delete [] d1;
	    delete [] d2;
	}
}

//--------------------------------------------------------------------------------------------------------
// add adge between vertices 'a' and 'b', if 'a' and 'b' have only one common neighbour
// input:  graph - cartesian product
//         graph2 - result of function RELATION_djoko_winkler()
// output: graph2 - new graph
//--------------------------------------------------------------------------------------------------------
void RelationTAU(myGraph &graph, myGraph &graph2)
//--------------------------------------------------------------------------------------------------------
{
	int icount = 0;
	
	myEdgeID edge_id = get(edge_index, graph);
		
	vIt viBegin, viEnd;
	aIt ai1, ai1_end, ai2, ai2_end, ai3, ai3_end, ai4, ai4_end;
	/*for(tie(viBegin, viEnd) = vertices(graph2); viBegin != viEnd; ++viBegin)
		cout << "\n v_nr: " << num_vertices(graph2) << " " << *viBegin; 
	*/
	for(tie(viBegin, viEnd) = vertices(graph); viBegin != viEnd; ++viBegin)
    {
    	for(tie(ai1, ai1_end) = adjacent_vertices(*viBegin, graph); ai1 != ai1_end; ++ai1)
      	{
      		for(tie(ai2, ai2_end) = adjacent_vertices(*ai1, graph); ai2 != ai2_end; ++ai2)
      		{
      			if(*ai2 != *viBegin)
      			{
      				for(tie(ai3, ai3_end) = adjacent_vertices(*ai2, graph); ai3 != ai3_end; ++ai3)
      				{
      					if((edge(*viBegin, *ai3, graph).second) && (*ai3 != *ai1))
      					{
      						icount++;
      					}
      				}
      				
     				if(icount == 0)
      				{   //cout << "\n1.test \n" << get(edge_id, edge(*viBegin, *ai1, graph).first) <<  " <-> " << get(edge_id, edge(*ai1, *ai2, graph).first) ;
      					if(!edge( get(edge_id, edge(*viBegin, *ai1, graph).first), get(edge_id, edge(*ai1, *ai2, graph).first), graph2).second)
      					{   //cout << "\n2.test \n" << get(edge_id, edge(*viBegin, *ai1, graph).first) <<  " <-> " << get(edge_id, edge(*ai1, *ai2, graph).first) ;  					
      						// edge(*viBegin, *ai1, graph).first => gibt die Koordinaten der Kante aus, die zw. den Knoten 'viBegin' und 'ai1' liegt
      						// get(edge_id, edge(*viBegin, *ai1, graph).first) => gibt die interne Kanten-ID der zuvor bestimmten Kante aus 
      						add_edge( get(edge_id, edge(*viBegin, *ai1, graph).first), get(edge_id, edge(*ai1, *ai2, graph).first), graph2);
      					}
      				}
    				icount = 0;
      			}
      		}
      	}
    }
}


//--------------------------------------------------------------------------------------------------------
// extract connected components of cartesian product based on BOOST-Library algorithm
// input:  graph - cartesian product
//         graph2 - result of function djoko_winklen and RelationTAU()
//         edgeVector - contained edges (myEdge) of cpgraph
// output: compositionen - vector of connected component elements of cartesian-product-graph
//--------------------------------------------------------------------------------------------------------
vector< vector<int> > decomposition(myGraph &graph, const myGraph &graph2 , const vector<myEdge> &edgeVector, vector<int>&  component, int& totalConnectComp)
//--------------------------------------------------------------------------------------------------------
{
  //  vector<int> component(num_vertices(graph2));
    vector<int> compVertex;
    vector< vector<int> > compositionen;
    /*int*/ //totalConnectComp = connected_components(graph2, &component[0]);
    eIt ei, ei_end;
    vIt vi, vi_end;
	
    vector<myGraph> graph_decomposition(totalConnectComp);
    
		
    for(int i = 0; i != totalConnectComp; i++)
	{
		for (tie(vi, vi_end) = vertices(graph); vi != vi_end; ++vi)
		{   
			//cout << "\nedge " << *vi << " in class " << component[*vi] ;  
	    	add_vertex(graph_decomposition.at(i));
	  	}
    }
	
    
	for(tie(vi, vi_end) = vertices(graph2); vi != vi_end; ++vi)
	{	
		int i=component[*vi];
		add_edge(edgeVector.at(*vi).first, edgeVector.at(*vi).second, graph_decomposition.at(i));
	}
	
	/*for(int i = 0 ; i < totalConnectComp ; i++){
		cout << "\ngraph " << i << endl;
		print_graph(graph_decomposition[i]);
	}*/
   // cout << "\n\n\n>>>> connected components of cartesian product <<<<" << endl << endl;
      
    /** einzelnen representanten rausfiltern: */
   //cout << ">> total connected_components CC(n) = " << totalConnectComp << endl;
   
    
  	
  	for(int i = 0; i < totalConnectComp; i++)
	{
		compVertex.clear();
		
		myGraph graphDecompositionAtI = graph_decomposition.at(i);
		vector<int> component_eq_cl(num_vertices(graphDecompositionAtI));
	 	/*int graphDecompConnectComp =*/ connected_components(graphDecompositionAtI, &component_eq_cl[0]);
	 	
		//cout << "\n>> number CC(" << i << ") of cartesian product = " << graphDecompConnectComp << endl;     
		//print_graph(component_eq_cl.at(0));		
		
		//cout << endl;
		for(tie(vi, vi_end) = vertices(graphDecompositionAtI); vi != vi_end; ++vi)
		{	 
			if(component_eq_cl[*vi] == 0) //cc)
			{
				compVertex.push_back(*vi);
				//cout << ">> comp " << component_eq_cl[*vi] << ": vertex " << *vi << endl;
			}
		}
			
		compositionen.push_back(compVertex);
		//}
    }
    return compositionen;
}


//--------------------------------------------------------------------------------------------------------
// extract single graph from cartesian-product-graph for test of isomorphism
// input:  graph - cartesian-product-graph
//         compositionen - vector of connected component elements of cartesian-product-graph
// output: graph - single graph for test of isomorphism
//
// clear_vertex(kntn, grph)  ==> entfernt alle Kanten vom Knoten 'kntn' des Graphens 'grph' und laesst den Knoten selbst uebrig
// remove_vertex(kntn, grph) ==> entfernt alle Knoten 'kntn' des Graphens 'grph' und laesst die Kanten uebrig
//--------------------------------------------------------------------------------------------------------
myGraph extractGraph(myGraph graph, vector<int> compositionen)
//--------------------------------------------------------------------------------------------------------
{
	vIt vi, vi_end, vi_temp;
	bool exist = false;
	int indexer = 0;
	int indexer_break = num_vertices(graph);
	
	//remove edges
	for(tie(vi, vi_end) = vertices(graph); vi != vi_end; ++vi)
	{
		for(unsigned int b = 0; b < compositionen.size(); b++)
		{	int t=*vi;		
			//if(*vi == compositionen.at(b))
			if(t == compositionen.at(b))
			{
				exist = true;
				break;
			}
			else
			{
				exist = false;
			}
		}
		
		if(exist == false)
		{
			clear_vertex(*vi, graph);
		}
	}
	
	//remove vertices without edges
	for(tie(vi, vi_end) = vertices(graph); vi != vi_end; ++vi)
	{
		indexer++;
		
		if(degree(*vi, graph) == 0)
		{
			remove_vertex(*vi, graph);
			--vi;
		}
		
		if(indexer == indexer_break)
			break;
	}
	
	return graph;
}



void decompose(myGraph& cpgraph, myGraph& graph2){
	
	//myGraph graph2/*=relation-graph*/;
	vector<myEdge> edgeVector;
	vector<myGraph> graphVector;
	vector<string> dirContent;
	vector< vector< vector<int> > > vecKomp;
	
	if( num_vertices(cpgraph) > 0)
	{
		//cout << "\n/** edges to vertices */";	
		edgeVector.clear();
		edgesToVertices(cpgraph, graph2, edgeVector);
		
		//cout << "\n/** relation Djoko-Winkler */";
		RELATION_djoko_winkler(cpgraph, graph2);
		
		//cout << "\n/** relation TAU */";
		RelationTAU(cpgraph, graph2);
		
		//cout << "\n/** decomposition of graphs */";
		/*vector<int> component(num_vertices(graph2));
		int totalConnectComp;
		totalConnectComp = connected_components(graph2, &component[0]);
		vector< vector<int> > compositionen = decomposition(cpgraph, graph2, edgeVector, component, totalConnectComp);*/
		//cout << "\n/** END decomposition of graphs */";
		
		//init_coordinates(cpgraph, graph2, component, totalConnectComp);
		
		/*ausgabe factors*/
	/*	cout << "\n\n\n>>> factors of cartesian skeleton: <<<" << endl;
		for(unsigned int t = 0; t < compositionen.size(); t++)
		{
			cout << "\n>> cartFactorNr " << (t + 1) << ": "  << endl;
			print_graph(extractGraph(cpgraph, compositionen.at(t)));
		}*/
		/*end_ausgabe factors*/
		
	}
}



	



